"""Annotations for checking imported modules within memory footprints"""

__all__ = ["ForbidImport", "ImportAnnotation", "RequireImport"]

import importlib
import sys

from abc import abstractmethod
from typing import Any, Dict, List

from .annotation import Annotation, AnnotationResult

from ..execution import MemoryFootprint


class ImportAnnotation(Annotation):
    """
    Annotation class for asserting conditions on the set of imported modules

    The set of modules imported in the student implementation is determined using ``sys.modules``,
    meaning that subclasses count any modules imported by third-party libaries, not just directly
    by the student's code.

    Args:
        module (``str``): the module name to forbid in importable form
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.annotations.annotation.Annotation>` constructor
    """

    module: str
    """the module name to forbid"""

    def __init__(self, module, **kwargs):
        if not isinstance(module, str):
            raise TypeError(f"{module} is not a string")

        try:
            already_imported = module in sys.modules
            importlib.import_module(module)
            if not already_imported:  # clean up sys.modules if needed
                sys.modules.pop(module)
        except Exception as e:
            raise ValueError(f"{module} is not importable")

        self.module = module

        super().__init__(**kwargs)

    @property
    def children(self) -> List[Annotation]:
        return []

    @abstractmethod
    def check(self, footprint: MemoryFootprint) -> AnnotationResult:
        ... # pragma: no cover

    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        To be equal to a ``ForbidImport``, the other object must also be a ``ForbidImport`` object, 
        have the same instance variables, and the same forbidden module.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.module == other.module

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation's details to a JSON-friendly dictionary format.

        Output dictionary contains the annotation's name, group, limit number, success message, and
        failure message, as well as the module name.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        d = super().to_dict()
        d.update({
            "module": str(self.module),
        })
        return d


class RequireImport(ImportAnnotation):
    """
    Annotation class for asserting that a specific module was imported.

    Args:
        module (``str``): the module name to forbid in importable form
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.annotations.annotation.Annotation>` constructor
    """

    def check(self, footprint: MemoryFootprint) -> AnnotationResult:
        """
        Checks that the memory footprint's imports contain the specified module.

        Args:
            footprint (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the
                memory footprint to check against
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation against ``footprint``
        """
        return AnnotationResult(self.module in footprint.imports, self)


class ForbidImport(ImportAnnotation):
    """
    Annotation class for asserting that a specific module was not imported.

    Args:
        module (``str``): the module name to forbid in importable form
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.annotations.annotation.Annotation>` constructor
    """

    def check(self, footprint: MemoryFootprint) -> AnnotationResult:
        """
        Checks that the memory footprint's imports don't contain the specified module.

        Args:
            footprint (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the
                memory footprint to check against
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation against ``footprint``
        """
        return AnnotationResult(self.module not in footprint.imports, self)



"""Annotations for marking-up reference implementations"""

from . import invariants
from .annotation import *
from .collection import *
from .complexity import *
from .import_ import *
from .initial_condition import *
from .relation import *
from .type_ import *
from .value import *



"""Utilities for PyBryt annotation tests"""

import numpy as np

from functools import lru_cache
from itertools import chain

from pybryt.execution import MemoryFootprint, MemoryFootprintValue


def assert_object_attrs(obj, attrs):
    """
    """
    for k, v in attrs.items():
        if k.endswith("__len"):
            assert len(getattr(obj, k[:-5])) == v, \
                f"Attr '{k}' length is wrong: expected {v} but got {len(getattr(obj, k[:-5]))}"
        else:
            attr = getattr(obj, k)
            if isinstance(attr, (np.ndarray, float, np.generic, int)):
                is_eq = np.allclose(attr, v)
            else:
                is_eq = attr == v
            if isinstance(is_eq, np.ndarray):
                assert is_eq.all(), f"Attr '{k}' is wrong: expected {v} but got {getattr(obj, k)}"
            else:
                assert is_eq, f"Attr '{k}' is wrong: expected {v} but got {getattr(obj, k)}"


@lru_cache(1)
def generate_memory_footprint() -> MemoryFootprint:
    """
    """
    np.random.seed(42)
    objs = [
        np.random.uniform(-100, 100, size=(100, 100)),
        4.0,
        list(range(100))[::-1],
        1,
        np.e,
        None,
        None,
        np.random.normal(size=102),
        4.0,
        "some CasE insenSITIve StrINg!",
    ]
    return MemoryFootprint.from_values(*[MemoryFootprintValue(o, i, None) for i, o in enumerate(objs)])



import setuptools
import sys

with open("README.md", "r") as fh:
	long_description = fh.read()

with open("requirements.txt") as f:
	install_requires = f.readlines()

# add dataclasses backport for python 3.6
if sys.version_info.minor < 7:
	install_requires.append("dataclasses")

# get version
env = {}
with open("pybryt/version.py") as f:
	exec(f.read(), env)
version = env["__version__"]

setuptools.setup(
	name = "pybryt",
	version = version,
	author = "Chris Pyles",
	author_email = "pybryt-support@microsoft.com",
	description = "Python auto-assessment library",
	long_description = long_description,
	long_description_content_type = "text/markdown",
	url = "https://github.com/microsoft/pybryt",
	license = "MIT",
	packages = setuptools.find_packages(exclude=["test"]),
	classifiers = [
		"Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
	],
	install_requires=install_requires,
	entry_points={
		"console_scripts": ["pybryt=pybryt.cli:cli"]
	},
)



"""Placeholders for student-set initial conditions"""

__all__ = ["InitialCondition"]

import operator

from typing import Any, Callable, Dict, List, Optional

from ..execution import MemoryFootprint


class InitialCondition:
    """
    A placeholder for initial conditions set in the submission.

    Args:
        name (``str``): the name of the initial condition
        transforms (``list[callable[[object], object]]``, optional): a list of transformations 
            (single-argument functions) to apply to the value in sequence
    """

    name: str
    """the name of the initial condition"""

    transforms: List[Callable[[Any], Any]]
    """transformations that should be applied to the value, in order"""

    def __init__(self, name: str, transforms: Optional[List[Callable[[Any], Any]]] = None):
        if not isinstance(name, str):
            raise TypeError("The name of an initial condition must be a string")

        self.name = name
        self.transforms = [] if transforms is None else transforms

    def apply(self, transform: Callable[[Any], Any]) -> "InitialCondition":
        """
        Apply an additional transformation to this initial condition, returning a new one.

        This instance is not modified; instead, a new initial condition instances is created with
        all transforms copied from this one but with the addition of the new transformation.

        Args:
            transform (``callable[[object], object]``): the transformation to add

        Returns:
            :py:class:`InitialCondition`: the new initial condition instance
        """
        transforms = self.transforms.copy()
        transforms.append(transform)
        return type(self)(self.name, transforms=transforms)

    def supply_footprint(self, footprint: MemoryFootprint) -> Any:
        """
        Calculate the value of this initial condition with all transformations applied based on the
        value stored in the provided memory footprint. 

        Args:
            initial_conditions (``dict[str, object]``): the initial conditions

        Returns:
            ``object``: the value after applying all transformations to it in sequence
        """
        return self.supply_values(footprint.get_initial_conditions())

    def supply_values(self, vals: Dict[str, Any]) -> Any:
        """
        Calculate the value of this initial condition with all transformations applied using the
        provided initial condition values.

        Applies all transformations in this initial condition to the supplied value and returns the
        result. If the result is itself an :py:class:`InitialCondition`, the values dictionary is
        supplied to that initial condition.

        Args:
            vals (``dict[str, object]``): the values to use as the initial conditions by name

        Returns:
            ``object``: the value after applying all transformations to it in sequence
        """
        if self.name not in vals:
            raise ValueError(f"The provided values do not have key '{self.name}'")

        ret = self.supply_value(vals[self.name])
        if isinstance(ret, type(self)):
            ret = ret.supply_values(vals)

        return ret

    def supply_value(self, val: Any) -> Any:
        """
        Calculate the value of this initial condition with all transformations applied.

        Applies all transformations in this initial condition to the supplied value and returns the
        result.

        Args:
            val (``object``): the value to use as the initial condition

        Returns:
            ``object``: the value after applying all transformations to it in sequence
        """
        for fn in self.transforms:
            val = fn(val)
        return val

    def _apply_binary_operator(
        self,
        op: Callable[[Any, Any], Any],
        other: Any,
        right: bool = False,
    ) -> "InitialCondition":
        """
        Create a new initial condition by applying a binary operator to this one.

        If ``right`` is true, then the value represented by this initial condition is passed as the
        second argument to the operator function; otherwise, it is passed as the first.

        Args:
            op (``callable[[object, object], object]``): a binary operator function
            other (``object``): the other object the operator should be applied to
            right (``bool``): whether this initial condition should be on the "right-side" of the
                operator

        Returns:
            :py:class:`InitialCondition`: the new initial condition instance
        """
        fn = lambda v: op(v, other)
        if right:
            fn = lambda v: op(other, v)

        return self.apply(fn)

    def __eq__(self, other) -> bool:
        """
        Determine whether the other object is also an initial condition with the same name and
        transforms.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return isinstance(other, type(self)) and self.name == other.name and \
            self.transforms == other.transforms

    def __add__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.add, other)

    def __radd__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.add, other, right=True)

    def __sub__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.sub, other)

    def __rsub__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.sub, other, right=True)

    def __mul__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.mul, other)

    def __rmul__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.mul, other, right=True)

    def __matmul__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.matmul, other)

    def __rmatmul__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.matmul, other, right=True)

    def __truediv__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.truediv, other)

    def __rtruediv__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.truediv, other, right=True)

    def __floordiv__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.floordiv, other)

    def __rfloordiv__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.floordiv, other, right=True)

    def __mod__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.mod, other)

    def __rmod__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.mod, other, right=True)

    def __divmod__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(divmod, other)

    def __rdivmod__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(divmod, other, right=True)

    def __pow__(self, other: Any, modulo: Any = None) -> "InitialCondition":
        if modulo is not None:  # support ternary pow
            return self.apply(lambda v: pow(v, other, modulo))
        return self._apply_binary_operator(operator.pow, other)

    def __rpow__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.pow, other, right=True)

    def __lshift__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.lshift, other)

    def __rlshift__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.lshift, other, right=True)

    def __rshift__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.rshift, other)

    def __rrshift__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.rshift, other, right=True)

    def __and__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.and_, other)

    def __rand__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.and_, other, right=True)

    def __xor__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.xor, other)

    def __rxor__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.xor, other, right=True)

    def __or__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.or_, other)

    def __ror__(self, other: Any) -> "InitialCondition":
        return self._apply_binary_operator(operator.or_, other, right=True)

    def __neg__(self) -> "InitialCondition":
        return self.apply(operator.neg)

    def __pos__(self) -> "InitialCondition":
        return self.apply(operator.pos)

    def __abs__(self) -> "InitialCondition":
        return self.apply(operator.abs)

    def __invert__(self) -> "InitialCondition":
        return self.apply(operator.invert)



""""""

from ..utils import AttrDict


def generate_mocked_frame(
    co_filename, co_name, f_lineno, f_globals={}, f_locals={}, f_trace=None, f_back=None
):
    """
    Factory for generating objects with the instance variables of an ``inspect`` frame that are 
    needed by PyBryt's trace function.
    """
    code = AttrDict({
        "co_filename": co_filename,
        "co_name": co_name,
    })
    return AttrDict({
        "f_lineno": f_lineno,
        "f_globals": f_globals,
        "f_locals": f_locals,
        "f_trace": f_trace,
        "f_back": f_back,
        "f_code": code,
    })


"""Tests for memory footprint objects"""

import nbformat
import pytest

from copy import deepcopy
from itertools import chain
from unittest import mock

from pybryt.execution.memory_footprint import (
    Counter, Event, MemoryFootprint, MemoryFootprintIterator, MemoryFootprintValue)


def generate_values():
    """
    Generate a list of memory footprint values for use in constructing memory footprints.

    Returns:
        ``list[MemoryFootprintValue]``: the values
    """
    return [MemoryFootprintValue(o, i + 1, None) for i, o in \
        enumerate([10, "", True, None, 10039, 1e-4])]


def test_counter():
    """
    Test the ``pybryt.execution.memory_footprint.Counter`` class.
    """
    counter = Counter()
    assert counter.get_value() == 0
    counter.increment()
    assert counter.get_value() == 1
    offset = 10
    counter.offset(offset)
    assert counter.get_value() == 1 + offset


def test_memory_footprint_constructor():
    """
    Test the constructor for ``pybryt.execution.memory_footprint.MemoryFootprint``.
    """
    footprint = MemoryFootprint()
    assert footprint.counter is not None
    assert len(footprint) == 0
    assert len(footprint.calls) == 0
    assert len(footprint.imports) == 0

    counter = Counter()
    counter.offset(50)
    footprint = MemoryFootprint(counter)
    assert footprint.counter is counter
    assert len(footprint) == 0
    assert len(footprint.calls) == 0
    assert len(footprint.imports) == 0


def test_from_values():
    """
    Test the ``pybryt.execution.memory_footprint.MemoryFootprint.from_values`` method.
    """
    vals = generate_values()
    footprint = MemoryFootprint.from_values(*vals)
    assert list(footprint) == vals
    assert footprint.num_steps == len(vals)
    assert footprint.counter.get_value() == len(vals)

    # check errors
    with pytest.raises(TypeError):
        MemoryFootprint.from_values(MemoryFootprintValue(1, 1, None), 2)


def test_combine():
    """
    Test the ``pybryt.execution.memory_footprint.MemoryFootprint.combine`` method.
    """
    vals = generate_values()
    v1, v2 = vals[:3], vals[3:]
    v2.extend(v1)

    f1, f2 = MemoryFootprint.from_values(*v1), MemoryFootprint.from_values(*v2)
    footprint = MemoryFootprint.combine(f1, f2)

    sorted_fp_vals = sorted(list(footprint), key=lambda mfp_val: mfp_val.timestamp)
    assert list(map(lambda mfp_val: mfp_val.value, sorted_fp_vals)) == list(map(lambda mfp_val: mfp_val.value, vals))
    assert footprint.counter.get_value() == len(v1) + len(v2)


def test_counter_manipulation():
    """
    Test counter manipulations for the ``MemoryFootprint`` class.
    """
    footprint = MemoryFootprint()
    footprint.increment_counter()
    assert footprint.counter.get_value() == 1
    footprint.offset_counter(50)
    assert footprint.counter.get_value() == 51


def test_values():
    """
    Test value updates and accesses for the ``MemoryFootprint`` class.
    """
    footprint = MemoryFootprint()
    footprint.add_value(1)
    assert len(footprint) == 1
    assert footprint.get_value(0).value == 1
    assert footprint.get_value(0).timestamp == 0

    footprint.increment_counter()
    footprint.add_value(2)
    assert len(footprint) == 2
    assert footprint.get_value(1).value == 2
    assert footprint.get_value(1).timestamp == 1
    assert footprint.get_value(1).event is None

    footprint.add_value(3, 10)
    assert len(footprint) == 3
    assert footprint.get_value(2).value == 3
    assert footprint.get_value(2).timestamp == 10

    footprint.increment_counter()
    footprint.add_value(2, event=Event.LINE)
    assert len(footprint) == 3
    assert footprint.get_value(1).event == Event.LINE

    footprint.add_value(2)
    assert len(footprint) == 3
    assert footprint.get_value(1).event == Event.LINE

    footprint.add_value(2, event=Event.RETURN)
    assert len(footprint) == 3
    assert footprint.get_value(1).event == Event.LINE_AND_RETURN


def test_calls():
    """
    Test call tracking for the ``MemoryFootprint`` class.
    """
    footprint = MemoryFootprint()
    footprint.add_call("foo", "bar")
    assert footprint.calls == [("foo", "bar")]


def test_imports():
    """
    Test import tracking for the ``MemoryFootprint`` class.
    """
    footprint = MemoryFootprint()
    imports = {"a", "b", "c", "d"}
    footprint.add_imports(*imports)
    assert footprint.imports == imports

    footprint.add_imports("a")
    assert footprint.imports == imports
    assert sum(i == "a" for i in footprint.imports) == 1


def test_set_executed_notebook():
    """
    Test executed notebook tracking for the ``MemoryFootprint`` class.
    """
    footprint = MemoryFootprint()
    assert footprint.executed_notebook == None

    nb = nbformat.v4.new_notebook()
    footprint.set_executed_notebook(nb)
    assert footprint.executed_notebook is nb


def test_filter_out_unpickleable_values():
    """
    Test pickleable value filtering tracking for the ``MemoryFootprint`` class.
    """
    vals = generate_values()
    footprint = MemoryFootprint.from_values(*vals)

    with mock.patch("pybryt.execution.memory_footprint.filter_pickleable_list") as mocked_filter:
        footprint.filter_out_unpickleable_values()
        mocked_filter.assert_called_once_with(footprint.values)


def test_eq():
    """
    Test ``MemoryFootprint`` equals comparisons.
    """
    vals = generate_values()
    f1, f2 = MemoryFootprint.from_values(*vals), MemoryFootprint()
    f3 = deepcopy(f1)

    assert f1 != f2
    assert f1 == f3


def test_misc_dunder_methods():
    """
    Tests for misc. dunder methods of ``MemoryFootprint``.
    """
    vals = generate_values()
    fp = MemoryFootprint.from_values(*vals)

    assert len(fp) == len(vals)

    fpi = iter(fp)
    assert isinstance(fpi, MemoryFootprintIterator)
    for e, a in zip(vals, fp):
        assert isinstance(a, MemoryFootprintValue)
        assert e == a


def test_initial_conditions():
    """
    Tests for initial conditions-related methods of ``MemoryFootprint``.
    """
    fp = MemoryFootprint()
    ics = {"foo": 1, "bar": 2}

    assert fp.get_initial_conditions() == {}

    fp.set_initial_conditions(ics)
    assert fp.get_initial_conditions() == ics

    new_ics = {"bar": 3, "baz": 4}
    ics.update(new_ics)

    fp.set_initial_conditions(new_ics)
    assert fp.get_initial_conditions() == ics



"""Time complexity analysis utilities"""

import inspect

from collections.abc import Sized
from itertools import chain
from typing import List, Optional, Union

from pybryt.execution.memory_footprint import MemoryFootprintValue

from .annotations import complexities as cplx, TimeComplexity
from .execution import (
    check_time_complexity, 
    get_tracing_frame,
    FrameTracer,
    MemoryFootprint,
    TimeComplexityResult,
)


ANNOTATION_NAME = "__time_complexity_checker__"


class TimeComplexityChecker:
    """
    A utility class for using PyBryt's tracing functionality to check the time complexity of a block
    of code.

    Uses PyBryt's tracing internals to set a trace function that counts the number of steps taken
    to execute a block of code, and then uses the complexity annotation framework to determine the
    best-matched complexity class of the block.
    """

    name: str
    """the name to use for the annotation"""

    results: List[TimeComplexityResult]
    """the result objects holding the step data for each input length"""

    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name if name is not None else ANNOTATION_NAME
        self.results = []

    def __call__(self, n: Union[int, float, Sized]) -> "_check_time_complexity_wrapper":
        """
        Create a wrapper for :py:class:`pybryt.execution.complexity.check_time_complexity` that enables
        tracing and collects the results object produced by that context manager.

        Args:
            n (``Union[int, float, Sized]``): the input length or the input itself if it supports ``len``

        Returns:
            :py:class:`_check_time_complexity_wrapper`: the initialized context manager that wraps
            ``check_time_complexity``
        """
        return _check_time_complexity_wrapper(self, n)

    def add_result(self, result: TimeComplexityResult) -> None:
        """
        Add a time complexity result to the collection of results.

        Args:
            result (``TimeComplexityResult``): the result object
        """
        self.results.append(result)

    def determine_complexity(self) -> cplx.complexity:
        """
        Determine the best-matched complexity class based on the results collected.

        Returns:
            :py:class:`pybryt.annotations.complexity.complexities.complexity`: the complexity class
            corresponding to the best-matched complexity
        """
        annot = TimeComplexity(cplx.constant, name=self.name)
        footprint = MemoryFootprint.from_values(
            *[MemoryFootprintValue(tcr, i, None) for i, tcr in enumerate(self.results)])
        result = annot.check(footprint)
        return result.value


class _check_time_complexity_wrapper:
    """
    A wrapper for :py:class:`pybryt.execution.complexity.check_time_complexity` that enables tracing
    for and collects the results from that context manager.

    Args:
        checker (:py:class:`TimeComplexityChecker`): the complexity checker that is using this wrapper
        n (``Union[int, float, Sized]``): the input length or the input itself if it supports ``len``
    """

    checker: TimeComplexityChecker
    """the complexity checker that is using this wrapper"""

    check_context: check_time_complexity
    """the time complexity context being wrapped"""

    frame_tracer: FrameTracer
    """the frame tracer being used to manage execution tracing"""

    n: Union[int, float, Sized]
    """the input length or the input itself if it supports ``len``"""

    def __init__(self, checker: TimeComplexityChecker, n: Union[int, float, Sized]) -> None:
        self.checker = checker
        self.n = n
        self.check_context = None
        self.frame_tracer = None

    def __enter__(self) -> None:
        self.frame_tracer = FrameTracer(inspect.currentframe().f_back)
        self.frame_tracer.start_trace()
        self.check_context = check_time_complexity(self.checker.name, self.n)
        self.check_context.__enter__()

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        self.check_context.__exit__(exc_type, exc_value, traceback)
        self.frame_tracer.end_trace()

        result = None
        for mfp_val in self.frame_tracer.get_footprint():
            if isinstance(mfp_val.value, TimeComplexityResult):
                result = mfp_val.value

        if result:
            self.checker.add_result(result)

        return False



"""Tests for complexity checking internals"""

import numpy as np
import pytest

import pybryt
import pybryt.execution

from .utils import generate_mocked_frame


def test_time_complexity():
    def do_work(n, transform):
        n = int(transform(n))
        for _ in range(n):
            2 * 2
        return n

    footprint, cir = pybryt.execution.create_collector()

    for e in range(1, 9):
        n = 10 ** e
        with pybryt.check_time_complexity("foo", n):
            do_work(n, lambda v: v)

    assert len(footprint) == 8
    assert all(isinstance(mfp_val.value, pybryt.TimeComplexityResult) for mfp_val in footprint)

    footprint.clear()

    for e in range(1, 9):
        n = float(10 ** e)
        with pybryt.check_time_complexity("foo", n):
            do_work(n, np.log2)

        n = np.random.uniform(size=int(n))
        with pybryt.check_time_complexity("bar", n):
            do_work(n, len)

    assert len(footprint) == 16
    assert all(isinstance(mfp_val.value, pybryt.TimeComplexityResult) for mfp_val in footprint)
    assert [mfp_val.value.name for mfp_val in footprint] == ["foo", "bar"] * 8

    # test error
    n = lambda v: v
    with pytest.raises(TypeError, match=f"n has invalid type {type(n)}"):
        with pybryt.check_time_complexity("foo", n):
            pass

    # check that tracking is disabled
    frame = generate_mocked_frame("<ipython-abc-123>", "bar", 1)
    with pybryt.check_time_complexity("foo", 10):
        cir(frame, "return", 10)
        assert len(footprint) == 16



"""Reference implementations for PyBryt"""

__all__ = ["ReferenceImplementation", "ReferenceResult", "generate_report"]

import nbformat
import numpy as np
import os
import warnings

from copy import deepcopy
from textwrap import indent
from typing import Any, Dict, List, Optional, Tuple, Union

from .annotations import Annotation, AnnotationResult
from .execution import MemoryFootprint
from .utils import get_stem, notebook_to_string, Serializable


class ReferenceImplementation(Serializable):
    """
    A reference implementation class for managing collections of annotations. Defines methods for
    creating, running, and storing reference implementations.

    Args:
        name (``str``): the name of the reference implementation
        annotations (``list[Annotation]``): the annotations comprising this reference implementation
    """
    
    annotations: List[Annotation]
    """the annotations comprising this reference implementation"""

    name: str
    """the name of the reference implementation"""

    display_name: Optional[str]
    """a name to use for this reference in text reports generated by PyBryt"""

    def __init__(self, name: str, annotations: List[Annotation], display_name: Optional[str] = None):
        if not isinstance(annotations, list):
            raise TypeError("annotations should be a list of Annotations")
        if not all(isinstance(ann, Annotation) for ann in annotations):
            raise TypeError("Found non-annotation in annotations")
        
        self.annotations, name_counts = [], {}
        for ann in annotations:
            track = True
            if ann.name is not None:
                count = name_counts.get(ann.name, 0)
                if ann.limit is None or ann.limit > count:
                    name_counts[ann.name] = count + 1
                else:
                    track = False

            if track:
                self.annotations.append(ann)

        self.name = name
        self.display_name = display_name

    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this reference implementation is equal to another object.

        For an object to equal a reference implementation, it must also be a reference 
        implementation and have the same annotations as this reference implementation.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return isinstance(other, type(self)) and self.annotations == other.annotations and \
            self.name == other.name and self.display_name == other.display_name

    @property
    def _default_dump_dest(self) -> str:
        return f"{self.name}.pkl"

    def get(self, name: str) -> Union[Annotation, List[Annotation]]:
        """
        Retrieves and returns annotation(s) using their ``name`` attribute.

        Returns a single annotation if there is only one annotation with that name, otherwise returns
        a list of annotations with that name.

        Args:
            name (``str``): the name of look up

        Returns:
            ``Annotation`` or ``list[Annotation]``: the annotation(s) with that name

        Raises:
            ``ValueError``: if there are no annotations with that name
        """
        annots = []
        for ann in self.annotations:
            if ann.name == name:
                annots.append(ann)
        if len(annots) == 0:
            raise ValueError(f"Found no annotations with name '{name}'")
        elif len(annots) == 1:
            return annots[0]
        else:
            return annots

    def run(self, footprint: MemoryFootprint, group: Optional[str] = None) -> 'ReferenceResult':        
        """
        Runs the annotations tracked by this reference implementation against a memory footprint.

        Can run only specific annotations by specifying the ``group`` argument. Returns a 
        :py:class:`ReferenceResult<pybryt.ReferenceResult>` object.

        Args:
            footprint (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the
                memory footprint to check against
            group (``str``, optional): if specified, only annotations in this group will be run

        Returns:
            :py:class:`ReferenceResult<pybryt.ReferenceResult>`: the results of this check

        Raises:
            ``ValueError``: if ``group`` is specified but there are no annotations with that group
        """
        if group is not None:
            annots = []
            for ann in self.annotations:
                if ann.group == group:
                    annots.append(ann)
            if len(annots) == 0:
                raise ValueError(f"Group '{group}' not found")
        
        else:
            annots = self.annotations
        
        results = []
        for exp in annots:
            results.append(exp.check(footprint))
        
        return ReferenceResult(self, results, group=group)

    @classmethod
    def compile(
        cls,
        path_or_nb: Union[str, nbformat.NotebookNode],
        **kwargs,
    ) -> Union['ReferenceImplementation', List['ReferenceImplementation']]:
        """
        Compiles a notebook or Python script into a single or list of reference implementations.

        Creates a reference implementation by executing a Jupyter Notebook or Python script and 
        collecting all :py:class:`ReferenceImplementation<pybryt.ReferenceImplementation>` objects
        created, if any, or all :py:class:`Annotation<pybryt.Annotation>` objects otherwise.

        Args:
            path_or_nb (``str`` or ``nbformat.NotebookNode``): the file to be executed
            **kwargs: additional keyword arguments passed to constructor
        
        Returns:
            ``ReferenceImplementation`` or ``list[ReferenceImplementation]``: the reference(s) 
            created by executing the file
        """
        Annotation.reset_tracked_annotations()

        if isinstance(path_or_nb, str) and os.path.splitext(path_or_nb)[1] != ".ipynb":
            with open(path_or_nb) as f:
                source = f.read()
        else:
            source = notebook_to_string(path_or_nb)

        if isinstance(path_or_nb, str) and "name" not in kwargs:
            kwargs["name"] = get_stem(path_or_nb)

        env = {}
        exec(source, env)

        refs = []
        for _, v in env.items():
            if isinstance(v, cls):
                refs.append(v)

        if not refs:
            if not Annotation.get_tracked_annotations():
                warnings.warn(f"Could not find any reference implementations in " \
                    f"{path_or_nb if isinstance(path_or_nb, str) else 'the provided notebook'}")
            else:
                name = kwargs.pop("name", None)
                if name is None:
                    raise ValueError("No name specified for the reference being compiled")
                refs = [cls(name, deepcopy(Annotation.get_tracked_annotations()), **kwargs)]

        Annotation.reset_tracked_annotations()
        
        if len(refs) == 1:
            return refs[0]
        return refs

    def get_name_for_report(self) -> str:
        """
        Get the name to use for the report in this reference.

        Returns:
            ``str``: the name
        """
        return self.display_name if self.display_name is not None else self.name


class ReferenceResult(Serializable):
    """
    Class for wrangling and managing the results of a reference implementation. Collects a series of
    :py:class:`AnnotationResult<pybryt.AnnotationResult>` objects and provides an API for managing
    those results collectively.

    Args:
        reference (``ReferenceImplementation``): the reference implementation
        annotation_results (``list[AnnotationResult]``): the annotation results from running the
            reference implementation
        group (``str``, optional): the name of the group of annotations executed, if applicable
    """

    reference: ReferenceImplementation
    """the reference implementation executed"""

    results: List[AnnotationResult]
    """the annotation results from running the reference implementation"""

    group: Optional[str]
    """the name of the group of annotations executed, if applicable"""

    def __init__(
        self, reference: ReferenceImplementation, annotation_results: List[AnnotationResult], 
        group: Optional[str] = None
    ):
        self.reference = reference
        self.results = annotation_results
        self.group = group

    def __repr__(self):
        results = ',\n  '.join(repr(r) for r in self.results)
        return f"ReferenceResult([\n  {results}\n])"

    @property
    def _default_dump_dest(self) -> str:
        return f"{self.name}_results.pkl"

    @property
    def correct(self) -> bool:
        """
        ``bool``: whether the reference implementation was satisfied
        """
        return all(r.satisfied for r in self.results)

    @property
    def name(self) -> str:
        """
        ``str``: the name of the underlying reference implementation
        """
        return self.reference.name

    @property
    def display_name(self) -> str:
        """
        ``str``: the display name of the underlying reference implementation
        """
        return self.reference.get_name_for_report()

    @property
    def messages(self) -> List[str]:
        """
        ``list[str]``: the list of messages returned by all annotations in the reference 
        implementation; if ``self.group`` is not ``None``, only messages from annotations in that
        group are returned
        """
        finalized_messages, message_indices_by_name = [], {}
        for r in self.results:
            for msg in r.messages:
                if msg.name in message_indices_by_name and not msg.satisfied:
                    idx = message_indices_by_name[msg.name]
                    finalized_messages[idx] = msg.message

                elif msg.name not in message_indices_by_name:
                    message_indices_by_name[msg.name] = len(finalized_messages)
                    finalized_messages.append(msg.message)

        return [m for m in finalized_messages if m is not None]

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this reference result's details to a JSON-friendly dictionary format.

        Output dictionary contains the group name run, if present, and the dictionary representations
        of all child annotation results.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        return {
            "group": self.group,
            "results": [ar.to_dict() for ar in self.results],
        }

    def to_array(self) -> np.ndarray:
        """
        Converts this result into a numpy array of integers, where ``1`` indicates a satisfied 
        annotation and ``0`` is unsatisfied.

        Returns:
            ``numpy.ndarray``: indicator array for the passage of annotations
        """
        return np.array([r.satisfied for r in self.results], dtype=int)


def generate_report(
    results: Union[ReferenceResult, List[ReferenceResult]], show_only: Optional[str] = None,
    fill_empty: bool = True
) -> str:
    """
    Collects a series of reference result objects and returns a summary of these results with any
    messages from the annotations.

    ``show_only`` should be in the set ``{'satisfied', 'unsatisfied', None}``. If ``"satisfied"``,
    the summary will contain only the results of satisfied reference implementations. If 
    ``"unsatisfied"``, the summary will contain only the results of unsatisfied reference 
    implementations. If ``None``, all reference implementations will be included.

    If ``show_only`` is set to a value that would result in an empty report (e.g. it is set to
    ``"satisfied"`` but no reference was satisfied) and ``fill_empty`` is ``True``, the report will
    be filled with any references that would normally be excluded by ``show_only``.

    Args:
        results (``Union[ReferenceResult, list[ReferenceResult]]``): the result(s) being collected
        show_only (``{'satisfied', 'unsatisfied', None}``): which results to report
        fill_empty (``bool``): if the resulting report would be empty, include results of the type
            not specified by ``show_only``

    Returns:
        ``str``: the summary report
    """
    if isinstance(results, ReferenceResult):
        results = [results]
    if not isinstance(results, list) or not all(isinstance(r, ReferenceResult) for r in results):
        raise TypeError("Cannot generate a report from arguments that are not reference result objects")
    if show_only not in {"satisfied", "unsatisfied", None}:
        raise ValueError("show_only must be in {'satisfied', 'unsatisfied', None}")
    
    filtered = []
    for res in results:
        if res.correct and show_only != "unsatisfied":
            filtered.append(res)
        elif not res.correct and show_only != "satisfied":
            filtered.append(res)

    if len(filtered) == 0 and fill_empty:
        filtered = results

    report = ""
    for i, res in enumerate(filtered):
        report += f"REFERENCE: {res.display_name}\n"
        if res.group is not None:
            report += f"GROUP: {res.group}\n"
        report += f"SATISFIED: {res.correct}"

        if res.messages:
            report += f"\nMESSAGES:\n"
            report += indent("\n".join(res.messages), "  - ")

        if i != len(filtered) - 1:
            report += "\n\n"

    return report
